package sqlite

import (
	"database/sql"
	"errors"
	"fmt"
	"log"

	"git.xdrm.io/go/articles-api/pkg/model"
	"git.xdrm.io/go/articles-api/pkg/persistence/storage"
	"github.com/jmoiron/sqlx"

	// sqlite3 driver
	_ "github.com/mattn/go-sqlite3"
)

// ErrNotInitialized is fired when using a non initialized database
var ErrNotInitialized = errors.New("storage not initialized")

// Storage using sqlite3 driver
type Storage struct {
	db *sqlx.DB
}

const schema = `
CREATE TABLE IF NOT EXISTS user(
	user_id INTEGER,
	username VARCHAR(100) NOT NULL,
	firstname VARCHAR(200) NOT NULL,
	lastname VARCHAR(200) NOT NULL,

	PRIMARY KEY(user_id)
);
CREATE TABLE IF NOT EXISTS article(
	article_id INTEGER,
	title VARCHAR(255) NOT NULL,
	body text NOT NULL,
	author_ref INTEGER,

	PRIMARY KEY(article_id),
	FOREIGN KEY(author_ref) REFERENCES user(user_id)
);
CREATE TABLE IF NOT EXISTS vote(
	user_ref INTEGER,
	article_ref VARCHAR(100) NOT NULL,
	value INTEGER NOT NULL,

	PRIMARY KEY(user_ref, article_ref),
	FOREIGN KEY(user_ref) REFERENCES user(user_id),
	FOREIGN KEY(article_ref) REFERENCES article(article_id),
	CHECK( value = 1 OR value = -1 )
);
`

// Open the database
func (s *Storage) Open() error {
	db, err := sqlx.Connect("sqlite3", "./local.db")
	if err != nil {
		return fmt.Errorf("cannot open storage: %w", err)
	}
	s.db = db

	// create tables
	tx, err := s.db.Begin()
	if err != nil {
		return fmt.Errorf("cannot create schema: %w", storage.Transaction)
	}
	_, err = tx.Exec(schema)
	if err != nil {
		return fmt.Errorf("cannot create tables: %s", err)
	}
	tx.Commit()
	return nil
}

// Close the databse
func (s *Storage) Close() error {
	return s.db.Close()
}

// CreateUser ...
func (s *Storage) CreateUser(username, firstname, lastname string) (*model.User, error) {
	user := model.User{
		Username:  username,
		Firstname: firstname,
		Lastname:  lastname,
	}

	// insert user
	tx, err := s.db.Beginx()
	if err != nil {
		return nil, storage.Transaction
	}
	res, err := tx.NamedExec(`INSERT INTO
		user(username, firstname, lastname)
		VALUES(:username,:firstname,:lastname);`, &user)
	if err != nil {
		return nil, storage.NotCreated
	}

	// find inserted id
	insertedID, err := res.LastInsertId()
	if err != nil {
		return nil, storage.NotFound
	}
	tx.Commit()
	user.ID = uint(insertedID)

	// return user info
	return &user, nil
}

// UpdateUser ...
func (s *Storage) UpdateUser(id uint, username, firstname, lastname *string) (*model.User, error) {
	tx, err := s.db.Beginx()
	if err != nil {
		return nil, storage.Transaction
	}

	// update fields
	if username != nil {
		_, err := tx.Exec(`UPDATE user SET username = ? WHERE user_id = ?;`, *username, id)
		if err != nil {
			return nil, storage.NotUpdated
		}

	}
	if firstname != nil {
		_, err := tx.Exec(`UPDATE user SET firstname = ? WHERE user_id = ?;`, *firstname, id)
		if err != nil {
			return nil, storage.NotUpdated
		}
	}
	if lastname != nil {
		_, err := tx.Exec(`UPDATE user SET lastname = ? WHERE user_id = ?;`, *lastname, id)
		if err != nil {
			return nil, storage.NotUpdated
		}
	}

	// select updated user
	user := model.User{}
	err = tx.Get(&user, `SELECT user_id, username, firstname, lastname FROM user WHERE user_id = $1;`, id)
	if err != nil {
		return nil, storage.NotFound
	}
	tx.Commit()

	return &user, nil
}

// DeleteUser ...
func (s *Storage) DeleteUser(id uint) error {
	tx, err := s.db.Beginx()
	if err != nil {
		return storage.Transaction
	}

	res, err := tx.Exec("DELETE FROM user WHERE user_id = ?;")
	if err != nil {
		return storage.NotDeleted
	}

	mustBeOne, err := res.RowsAffected()
	if err != nil || mustBeOne < 1 {
		return storage.NotDeleted
	}
	tx.Commit()

	return nil
}

// GetUserByID ...
func (s *Storage) GetUserByID(id uint) (*model.User, error) {
	user := model.User{}

	err := s.db.Get(&user, `SELECT user_id, username, firstname, lastname FROM user WHERE user_id = $1;`, id)
	if err != nil {
		if err == sql.ErrNoRows {
			return nil, storage.NotFound
		}
		return nil, storage.Failure
	}

	return &user, nil
}

// GetAllUsers ...
func (s *Storage) GetAllUsers() ([]model.User, error) {
	users := []model.User{}
	err := s.db.Select(&users, "SELECT user_id, username, firstname, lastname FROM user;")
	if err != nil {
		return nil, storage.NotFound
	}
	return users, nil
}

// CreateArticle ...
func (s *Storage) CreateArticle(title, body string, author uint) (*model.Article, error) {
	article := model.Article{
		Title:  title,
		Body:   body,
		Author: author,
	}

	// insert article
	tx, err := s.db.Beginx()
	if err != nil {
		return nil, storage.Transaction
	}

	res, err := tx.NamedExec(`INSERT INTO
		article(title, body, author_ref)
		VALUES(:title,:body,:author_ref);`, &article)
	if err != nil {
		log.Printf("err: %s", err)
		return nil, storage.NotCreated
	}

	// find inserted id
	insertedID, err := res.LastInsertId()
	if err != nil {
		return nil, storage.NotFound
	}
	tx.Commit()
	article.ID = uint(insertedID)

	return &article, nil
}

// UpdateArticle ...
func (s *Storage) UpdateArticle(id uint, title *string, body *string) (*model.Article, error) {
	tx, err := s.db.Beginx()
	if err != nil {
		return nil, storage.Transaction
	}

	if title != nil {
		_, err := tx.Exec(`UPDATE article SET title = ? WHERE article_id = ?;`, *title, id)
		if err != nil {
			return nil, storage.NotUpdated
		}
	}
	if body != nil {
		_, err := tx.Exec(`UPDATE article SET body = ? WHERE article_id = ?;`, *body, id)
		if err != nil {
			return nil, storage.NotUpdated
		}
	}

	// select updated article
	article, err := s.GetArticleByID(id)
	if err != nil {
		return nil, storage.NotFound
	}
	tx.Commit()

	return article, nil
}

// DeleteArticle ...
func (s *Storage) DeleteArticle(id uint) error {
	tx, err := s.db.Beginx()
	if err != nil {
		return storage.Transaction
	}

	res, err := tx.Exec("DELETE FROM article WHERE article_id = ?;")
	if err != nil {
		return storage.NotDeleted
	}

	mustBeOne, err := res.RowsAffected()
	if err != nil || mustBeOne < 1 {
		return storage.NotDeleted
	}
	tx.Commit()

	return nil
}

// GetArticleByID ...
func (s *Storage) GetArticleByID(id uint) (*model.Article, error) {
	article := model.Article{}

	err := s.db.Get(&article, `SELECT article_id, title, body, author_ref, COALESCE(SUM(vote.value),0) as score
		FROM article
		LEFT OUTER JOIN vote ON article.article_id = vote.article_ref
		WHERE article_id = $1
		GROUP BY article_id, title, body, author_ref
		LIMIT 1;`, id)
	if err != nil {
		log.Printf("err: %s", err)
		return nil, storage.NotFound
	}

	return &article, nil
}

// GetArticlesByAuthor ...
func (s *Storage) GetArticlesByAuthor(id uint) ([]model.Article, error) {
	articles := []model.Article{}

	err := s.db.Select(&articles, `SELECT article_id, title, body, author_ref, COALESCE(SUM(vote.value),0) as score
		FROM article
		LEFT OUTER JOIN vote ON article.article_id = vote.article_ref
		WHERE author_ref = $1
		GROUP BY article_id, title, body, author_ref;`, id)
	if err != nil {
		log.Printf("err: %s", err)
		return nil, storage.NotFound
	}

	return articles, nil
}

// GetAllArticles ...
func (s *Storage) GetAllArticles() ([]model.Article, error) {
	articles := []model.Article{}

	err := s.db.Select(&articles, `SELECT article_id, title, body, author_ref, COALESCE(SUM(vote.value),0) as score
		FROM article
		LEFT OUTER JOIN vote ON article.article_id = vote.article_ref
		GROUP BY article_id, title, body, author_ref;`)
	if err != nil {
		return nil, storage.NotFound
	}

	return articles, nil
}

// UpVote ...
func (s *Storage) UpVote(user uint, article uint) (*model.Vote, error) {
	// update if alreay exists
	vote := model.Vote{
		User:    user,
		Article: article,
		Value:   1,
	}

	tx, err := s.db.Beginx()
	if err != nil {
		return nil, storage.Transaction
	}

	// alreay exists -> update
	row := tx.QueryRow("SELECT user_ref, article_ref, value FROM vote WHERE article_ref=? AND user_ref=?;", vote.Article, vote.User)
	if row != nil && row.Scan() != sql.ErrNoRows {
		_, err := tx.Exec(`UPDATE vote SET value = ? WHERE article_ref = ? AND user_ref = ?;`, vote.Value, vote.Article, vote.User)
		if err != nil {
			return nil, storage.NotUpdated
		}

		tx.Commit()
		return &vote, nil
	}

	// create vote
	_, err = tx.Exec("INSERT INTO vote(user_ref, article_ref, value) VALUES(?, ?, ?);", vote.User, vote.Article, vote.Value)
	if err != nil {
		return nil, storage.NotCreated
	}
	tx.Commit()

	return &vote, nil
}

// DownVote ...
func (s *Storage) DownVote(user uint, article uint) (*model.Vote, error) {
	// update if alreay exists
	vote := model.Vote{
		User:    user,
		Article: article,
		Value:   -1,
	}

	tx, err := s.db.Beginx()
	if err != nil {
		return nil, storage.Transaction
	}

	// alreay exists -> update
	row := tx.QueryRow("SELECT user_ref, article_ref, value FROM vote WHERE article_ref=? AND user_ref=?;", vote.Article, vote.User)
	if row != nil && row.Scan() != sql.ErrNoRows {
		_, err := tx.Exec(`UPDATE vote SET value = ? WHERE article_ref = ? AND user_ref = ?;`, vote.Value, vote.Article, vote.User)
		if err != nil {
			return nil, storage.NotUpdated
		}

		tx.Commit()
		return &vote, nil
	}

	// create vote
	_, err = tx.Exec("INSERT INTO vote(user_ref, article_ref, value) VALUES(?, ?, ?);", vote.User, vote.Article, vote.Value)
	if err != nil {
		return nil, storage.NotCreated
	}

	tx.Commit()
	return &vote, nil
}
